#ifndef _CXS_H
#define _CXS_H
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#ifdef MATLAB_MEX_FILE
#include "mex.h"
#endif

#ifdef __cplusplus
extern "C" {
#else
#include <complex.h>
#endif

#define CS_VER 1		    /* CXSparse Version 1.2.0 */
#define CS_SUBVER 2
#define CS_SUBSUB 0
#define CS_DATE "Mar 6, 2006"	    /* CXSparse release date */
#define CS_COPYRIGHT "Copyright (c) Timothy A. Davis, 2006"
#define CXSPARSE

/* -------------------------------------------------------------------------- */
/* double/int version of CXSparse */
/* -------------------------------------------------------------------------- */

/* --- primary CSparse routines and data structures ------------------------- */

typedef struct cs_di_sparse  /* matrix in compressed-column or triplet form */
{
    int nzmax ;	    /* maximum number of entries */
    int m ;	    /* number of rows */
    int n ;	    /* number of columns */
    int *p ;	    /* column pointers (size n+1) or col indices (size nzmax) */
    int *i ;	    /* row indices, size nzmax */
    double *x ;	    /* numerical values, size nzmax */
    int nz ;	    /* # of entries in triplet matrix, -1 for compressed-col */
} cs_di ;

cs_di *cs_di_add (const cs_di *A, const cs_di *B, double alpha, double beta) ;
int cs_di_cholsol (const cs_di *A, double *b, int order) ;
int cs_di_dupl (cs_di *A) ;
int cs_di_entry (cs_di *T, int i, int j, double x) ;
int cs_di_lusol (const cs_di *A, double *b, int order, double tol) ;
int cs_di_gaxpy (const cs_di *A, const double *x, double *y) ;
cs_di *cs_di_multiply (const cs_di *A, const cs_di *B) ;
int cs_di_qrsol (const cs_di *A, double *b, int order) ;
cs_di *cs_di_transpose (const cs_di *A, int values) ;
cs_di *cs_di_triplet (const cs_di *T) ;
double cs_di_norm (const cs_di *A) ;
int cs_di_print (const cs_di *A, int brief) ;
cs_di *cs_di_load (FILE *f) ;

/* utilities */
void *cs_di_calloc (int n, size_t size) ;
void *cs_di_free (void *p) ;
void *cs_di_realloc (void *p, int n, size_t size, int *ok) ;
cs_di *cs_di_spalloc (int m, int n, int nzmax, int values, int t) ;
cs_di *cs_di_spfree (cs_di *A) ;
int cs_di_sprealloc (cs_di *A, int nzmax) ;
void *cs_di_malloc (int n, size_t size) ;

/* --- secondary CSparse routines and data structures ----------------------- */

typedef struct cs_di_symbolic  /* symbolic Cholesky, LU, or QR analysis */
{
    int *Pinv ;	    /* inverse row perm. for QR, fill red. perm for Chol */
    int *Q ;	    /* fill-reducing column permutation for LU and QR */
    int *parent ;   /* elimination tree for Cholesky and QR */
    int *cp ;	    /* column pointers for Cholesky, row counts for QR */
    int m2 ;	    /* # of rows for QR, after adding fictitious rows */
    int lnz ;	    /* # entries in L for LU or Cholesky; in V for QR */
    int unz ;	    /* # entries in U for LU; in R for QR */
} cs_dis ;

typedef struct cs_di_numeric   /* numeric Cholesky, LU, or QR factorization */
{
    cs_di *L ;	    /* L for LU and Cholesky, V for QR */
    cs_di *U ;	    /* U for LU, R for QR, not used for Cholesky */
    int *Pinv ;	    /* partial pivoting for LU */
    double *B ;	    /* beta [0..n-1] for QR */
} cs_din ;

typedef struct cs_di_dmperm_results    /* cs_di_dmperm or cs_di_scc output */
{
    int *P ;	    /* size m, row permutation */
    int *Q ;	    /* size n, column permutation */
    int *R ;	    /* size nb+1, block k is rows R[k] to R[k+1]-1 in A(P,Q) */
    int *S ;	    /* size nb+1, block k is cols S[k] to S[k+1]-1 in A(P,Q) */
    int nb ;	    /* # of blocks in fine dmperm decomposition */
    int rr [5] ;    /* coarse row decomposition */
    int cc [5] ;    /* coarse column decomposition */
} cs_did ;

int *cs_di_amd (const cs_di *A, int order) ;
cs_din *cs_di_chol (const cs_di *A, const cs_dis *S) ;
cs_did *cs_di_dmperm (const cs_di *A) ;
int cs_di_droptol (cs_di *A, double tol) ;
int cs_di_dropzeros (cs_di *A) ;
int cs_di_happly (const cs_di *V, int i, double beta, double *x) ;
int cs_di_ipvec (int n, const int *P, const double *b, double *x) ;
int cs_di_lsolve (const cs_di *L, double *x) ;
int cs_di_ltsolve (const cs_di *L, double *x) ;
cs_din *cs_di_lu (const cs_di *A, const cs_dis *S, double tol) ;
cs_di *cs_di_permute (const cs_di *A, const int *P, const int *Q, int values) ;
int *cs_di_pinv (const int *P, int n) ;
int cs_di_pvec (int n, const int *p, const double *b, double *x) ;
cs_din *cs_di_qr (const cs_di *A, const cs_dis *S) ;
cs_dis *cs_di_schol (const cs_di *A, int order) ;
cs_dis *cs_di_sqr (const cs_di *A, int order, int qr) ;
cs_di *cs_di_symperm (const cs_di *A, const int *Pinv, int values) ;
int cs_di_usolve (const cs_di *U, double *x) ;
int cs_di_utsolve (const cs_di *U, double *x) ;
int cs_di_updown (cs_di *L, int sigma, const cs_di *C, const int *parent) ;

/* utilities */
cs_dis *cs_di_sfree (cs_dis *S) ;
cs_din *cs_di_nfree (cs_din *N) ;
cs_did *cs_di_dfree (cs_did *D) ;

/* --- tertiary CSparse routines -------------------------------------------- */

int *cs_di_counts (const cs_di *A, const int *parent, const int *post,
    int ata) ;
int cs_di_cumsum (int *p, int *c, int n) ;
int cs_di_dfs (int j, cs_di *L, int top, int *xi, int *pstack,
    const int *Pinv) ;
int *cs_di_etree (const cs_di *A, int ata) ;
int cs_di_fkeep (cs_di *A, int (*fkeep) (int, int, double, void *),
    void *other) ;
double cs_di_house (double *x, double *beta, int n) ;
int *cs_di_maxtrans (const cs_di *A) ;
int *cs_di_post (int n, const int *parent) ;
cs_did *cs_di_scc (cs_di *A) ;
int cs_di_scatter (const cs_di *A, int j, double beta, int *w, double *x,
    int mark, cs_di *C, int nz) ;
int cs_di_tdfs (int j, int k, int *head, const int *next, int *post,
    int *stack) ;

/* utilities */
cs_did *cs_di_dalloc (int m, int n) ;
cs_di *cs_di_done (cs_di *C, void *w, void *x, int ok) ;
int *cs_di_idone (int *p, cs_di *C, void *w, int ok) ;
cs_din *cs_di_ndone (cs_din *N, cs_di *C, void *w, void *x, int ok) ;
cs_did *cs_di_ddone (cs_did *D, cs_di *C, void *w, int ok) ;

/* nearly private routines, only used in cs_di_lu */
int cs_di_reach (cs_di *L, const cs_di *B, int k, int *xi, const int *Pinv) ;
int cs_di_splsolve (cs_di *L, const cs_di *B, int k, int *xi, double *x,
    const int *Pinv) ;


/* -------------------------------------------------------------------------- */
/* double/long version of CXSparse */
/* -------------------------------------------------------------------------- */

/* --- primary CSparse routines and data structures ------------------------- */

typedef struct cs_dl_sparse  /* matrix in compressed-column or triplet form */
{
    long nzmax ;    /* maximum number of entries */
    long m ;	    /* number of rows */
    long n ;	    /* number of columns */
    long *p ;	    /* column pointers (size n+1) or col indlces (size nzmax) */
    long *i ;	    /* row indices, size nzmax */
    double *x ;	    /* numerical values, size nzmax */
    long nz ;	    /* # of entries in triplet matrix, -1 for compressed-col */
} cs_dl ;

cs_dl *cs_dl_add (const cs_dl *A, const cs_dl *B, double alpha, double beta) ;
long cs_dl_cholsol (const cs_dl *A, double *b, long order) ;
long cs_dl_dupl (cs_dl *A) ;
long cs_dl_entry (cs_dl *T, long i, long j, double x) ;
long cs_dl_lusol (const cs_dl *A, double *b, long order, double tol) ;
long cs_dl_gaxpy (const cs_dl *A, const double *x, double *y) ;
cs_dl *cs_dl_multiply (const cs_dl *A, const cs_dl *B) ;
long cs_dl_qrsol (const cs_dl *A, double *b, long order) ;
cs_dl *cs_dl_transpose (const cs_dl *A, long values) ;
cs_dl *cs_dl_triplet (const cs_dl *T) ;
double cs_dl_norm (const cs_dl *A) ;
long cs_dl_print (const cs_dl *A, long brief) ;
cs_dl *cs_dl_load (FILE *f) ;

/* utilities */
void *cs_dl_calloc (long n, size_t size) ;
void *cs_dl_free (void *p) ;
void *cs_dl_realloc (void *p, long n, size_t size, long *ok) ;
cs_dl *cs_dl_spalloc (long m, long n, long nzmax, long values, long t) ;
cs_dl *cs_dl_spfree (cs_dl *A) ;
long cs_dl_sprealloc (cs_dl *A, long nzmax) ;
void *cs_dl_malloc (long n, size_t size) ;

/* --- secondary CSparse routines and data structures ----------------------- */

typedef struct cs_dl_symbolic  /* symbolic Cholesky, LU, or QR analysis */
{
    long *Pinv ;    /* inverse row perm. for QR, fill red. perm for Chol */
    long *Q ;	    /* fill-reducing column permutation for LU and QR */
    long *parent ;  /* elimination tree for Cholesky and QR */
    long *cp ;	    /* column pointers for Cholesky, row counts for QR */
    long m2 ;	    /* # of rows for QR, after adding fictitious rows */
    long lnz ;	    /* # entries in L for LU or Cholesky; in V for QR */
    long unz ;	    /* # entries in U for LU; in R for QR */
} cs_dls ;

typedef struct cs_dl_numeric   /* numeric Cholesky, LU, or QR factorization */
{
    cs_dl *L ;	    /* L for LU and Cholesky, V for QR */
    cs_dl *U ;	    /* U for LU, R for QR, not used for Cholesky */
    long *Pinv ;    /* partial pivoting for LU */
    double *B ;	    /* beta [0..n-1] for QR */
} cs_dln ;

typedef struct cs_dl_dmperm_results    /* cs_dl_dmperm or cs_dl_scc output */
{
    long *P ;	    /* size m, row permutation */
    long *Q ;	    /* size n, column permutation */
    long *R ;	    /* size nb+1, block k is rows R[k] to R[k+1]-1 in A(P,Q) */
    long *S ;	    /* size nb+1, block k is cols S[k] to S[k+1]-1 in A(P,Q) */
    long nb ;	    /* # of blocks in fine dmperm decomposition */
    long rr [5] ;   /* coarse row decomposition */
    long cc [5] ;   /* coarse column decomposition */
} cs_dld ;

long *cs_dl_amd (const cs_dl *A, long order) ;
cs_dln *cs_dl_chol (const cs_dl *A, const cs_dls *S) ;
cs_dld *cs_dl_dmperm (const cs_dl *A) ;
long cs_dl_droptol (cs_dl *A, double tol) ;
long cs_dl_dropzeros (cs_dl *A) ;
long cs_dl_happly (const cs_dl *V, long i, double beta, double *x) ;
long cs_dl_ipvec (long n, const long *P, const double *b, double *x) ;
long cs_dl_lsolve (const cs_dl *L, double *x) ;
long cs_dl_ltsolve (const cs_dl *L, double *x) ;
cs_dln *cs_dl_lu (const cs_dl *A, const cs_dls *S, double tol) ;
cs_dl *cs_dl_permute (const cs_dl *A, const long *P, const long *Q,
    long values) ;
long *cs_dl_pinv (const long *P, long n) ;
long cs_dl_pvec (long n, const long *p, const double *b, double *x) ;
cs_dln *cs_dl_qr (const cs_dl *A, const cs_dls *S) ;
cs_dls *cs_dl_schol (const cs_dl *A, long order) ;
cs_dls *cs_dl_sqr (const cs_dl *A, long order, long qr) ;
cs_dl *cs_dl_symperm (const cs_dl *A, const long *Pinv, long values) ;
long cs_dl_usolve (const cs_dl *U, double *x) ;
long cs_dl_utsolve (const cs_dl *U, double *x) ;
long cs_dl_updown (cs_dl *L, long sigma, const cs_dl *C, const long *parent) ;

/* utilities */
cs_dls *cs_dl_sfree (cs_dls *S) ;
cs_dln *cs_dl_nfree (cs_dln *N) ;
cs_dld *cs_dl_dfree (cs_dld *D) ;

/* --- tertiary CSparse routines -------------------------------------------- */

long *cs_dl_counts (const cs_dl *A, const long *parent, const long *post,
    long ata) ;
long cs_dl_cumsum (long *p, long *c, long n) ;
long cs_dl_dfs (long j, cs_dl *L, long top, long *xi, long *pstack,
    const long *Pinv) ;
long *cs_dl_etree (const cs_dl *A, long ata) ;
long cs_dl_fkeep (cs_dl *A, long (*fkeep) (long, long, double, void *),
    void *other) ;
double cs_dl_house (double *x, double *beta, long n) ;
long *cs_dl_maxtrans (const cs_dl *A) ;
long *cs_dl_post (long n, const long *parent) ;
cs_dld *cs_dl_scc (cs_dl *A) ;
long cs_dl_scatter (const cs_dl *A, long j, double beta, long *w, double *x,
    long mark,cs_dl *C, long nz) ;
long cs_dl_tdfs (long j, long k, long *head, const long *next, long *post,
    long *stack) ;

/* utilities */
cs_dld *cs_dl_dalloc (long m, long n) ;
cs_dl *cs_dl_done (cs_dl *C, void *w, void *x, long ok) ;
long *cs_dl_idone (long *p, cs_dl *C, void *w, long ok) ;
cs_dln *cs_dl_ndone (cs_dln *N, cs_dl *C, void *w, void *x, long ok) ;
cs_dld *cs_dl_ddone (cs_dld *D, cs_dl *C, void *w, long ok) ;

/* nearly private routines, only used in cs_dl_lu */
long cs_dl_reach (cs_dl *L, const cs_dl *B, long k, long *xi,
    const long *Pinv) ;
long cs_dl_splsolve (cs_dl *L, const cs_dl *B, long k, long *xi, double *x,
    const long *Pinv) ;


/* -------------------------------------------------------------------------- */
/* complex/int version of CXSparse */
/* -------------------------------------------------------------------------- */

/* --- primary CSparse routines and data structures ------------------------- */

typedef struct cs_ci_sparse  /* matrix in compressed-column or triplet form */
{
    int nzmax ;	    /* maximum number of entries */
    int m ;	    /* number of rows */
    int n ;	    /* number of columns */
    int *p ;	    /* column pointers (size n+1) or col indices (size nzmax) */
    int *i ;	    /* row indices, size nzmax */
    double _Complex *x ;    /* numerical values, size nzmax */
    int nz ;	    /* # of entries in triplet matrix, -1 for compressed-col */
} cs_ci ;

cs_ci *cs_ci_add (const cs_ci *A, const cs_ci *B, double _Complex alpha,
    double _Complex beta) ;
int cs_ci_cholsol (const cs_ci *A, double _Complex *b, int order) ;
int cs_ci_dupl (cs_ci *A) ;
int cs_ci_entry (cs_ci *T, int i, int j, double _Complex x) ;
int cs_ci_lusol (const cs_ci *A, double _Complex *b, int order, double tol) ;
int cs_ci_gaxpy (const cs_ci *A, const double _Complex *x, double _Complex *y) ;
cs_ci *cs_ci_multiply (const cs_ci *A, const cs_ci *B) ;
int cs_ci_qrsol (const cs_ci *A, double _Complex *b, int order) ;
cs_ci *cs_ci_transpose (const cs_ci *A, int values) ;
cs_ci *cs_ci_triplet (const cs_ci *T) ;
double cs_ci_norm (const cs_ci *A) ;
int cs_ci_print (const cs_ci *A, int brief) ;
cs_ci *cs_ci_load (FILE *f) ;

/* utilities */
void *cs_ci_calloc (int n, size_t size) ;
void *cs_ci_free (void *p) ;
void *cs_ci_realloc (void *p, int n, size_t size, int *ok) ;
cs_ci *cs_ci_spalloc (int m, int n, int nzmax, int values, int t) ;
cs_ci *cs_ci_spfree (cs_ci *A) ;
int cs_ci_sprealloc (cs_ci *A, int nzmax) ;
void *cs_ci_malloc (int n, size_t size) ;

/* --- secondary CSparse routines and data structures ----------------------- */

typedef struct cs_ci_symbolic  /* symbolic Cholesky, LU, or QR analysis */
{
    int *Pinv ;	    /* inverse row perm. for QR, fill red. perm for Chol */
    int *Q ;	    /* fill-reducing column permutation for LU and QR */
    int *parent ;   /* elimination tree for Cholesky and QR */
    int *cp ;	    /* column pointers for Cholesky, row counts for QR */
    int m2 ;	    /* # of rows for QR, after adding fictitious rows */
    int lnz ;	    /* # entries in L for LU or Cholesky; in V for QR */
    int unz ;	    /* # entries in U for LU; in R for QR */
} cs_cis ;

typedef struct cs_ci_numeric   /* numeric Cholesky, LU, or QR factorization */
{
    cs_ci *L ;	    /* L for LU and Cholesky, V for QR */
    cs_ci *U ;	    /* U for LU, R for QR, not used for Cholesky */
    int *Pinv ;	    /* partial pivoting for LU */
    double _Complex *B ;	    /* beta [0..n-1] for QR */
} cs_cin ;

typedef struct cs_ci_dmperm_results    /* cs_ci_dmperm or cs_ci_scc output */
{
    int *P ;	    /* size m, row permutation */
    int *Q ;	    /* size n, column permutation */
    int *R ;	    /* size nb+1, block k is rows R[k] to R[k+1]-1 in A(P,Q) */
    int *S ;	    /* size nb+1, block k is cols S[k] to S[k+1]-1 in A(P,Q) */
    int nb ;	    /* # of blocks in fine dmperm decomposition */
    int rr [5] ;    /* coarse row decomposition */
    int cc [5] ;    /* coarse column decomposition */
} cs_cid ;

int *cs_ci_amd (const cs_ci *A, int order) ;
cs_cin *cs_ci_chol (const cs_ci *A, const cs_cis *S) ;
cs_cid *cs_ci_dmperm (const cs_ci *A) ;
int cs_ci_droptol (cs_ci *A, double tol) ;
int cs_ci_dropzeros (cs_ci *A) ;
int cs_ci_happly (const cs_ci *V, int i, double _Complex beta,
    double _Complex *x) ;
int cs_ci_ipvec (int n, const int *P, const double _Complex *b,
    double _Complex *x) ;
int cs_ci_lsolve (const cs_ci *L, double _Complex *x) ;
int cs_ci_ltsolve (const cs_ci *L, double _Complex *x) ;
cs_cin *cs_ci_lu (const cs_ci *A, const cs_cis *S, double tol) ;
cs_ci *cs_ci_permute (const cs_ci *A, const int *P, const int *Q, int values) ;
int *cs_ci_pinv (const int *P, int n) ;
int cs_ci_pvec (int n, const int *p, const double _Complex *b,
    double _Complex *x) ;
cs_cin *cs_ci_qr (const cs_ci *A, const cs_cis *S) ;
cs_cis *cs_ci_schol (const cs_ci *A, int order) ;
cs_cis *cs_ci_sqr (const cs_ci *A, int order, int qr) ;
cs_ci *cs_ci_symperm (const cs_ci *A, const int *Pinv, int values) ;
int cs_ci_usolve (const cs_ci *U, double _Complex *x) ;
int cs_ci_utsolve (const cs_ci *U, double _Complex *x) ;
int cs_ci_updown (cs_ci *L, int sigma, const cs_ci *C, const int *parent) ;

/* utilities */
cs_cis *cs_ci_sfree (cs_cis *S) ;
cs_cin *cs_ci_nfree (cs_cin *N) ;
cs_cid *cs_ci_dfree (cs_cid *D) ;

/* --- tertiary CSparse routines -------------------------------------------- */

int *cs_ci_counts (const cs_ci *A, const int *parent, const int *post,
    int ata) ;
int cs_ci_cumsum (int *p, int *c, int n) ;
int cs_ci_dfs (int j, cs_ci *L, int top, int *xi, int *pstack,
    const int *Pinv) ;
int *cs_ci_etree (const cs_ci *A, int ata) ;
int cs_ci_fkeep (cs_ci *A, int (*fkeep) (int, int, double _Complex, void *),
    void *other) ;
double _Complex cs_ci_house (double _Complex *x, double _Complex *beta, int n) ;
int *cs_ci_maxtrans (const cs_ci *A) ;
int *cs_ci_post (int n, const int *parent) ;
cs_cid *cs_ci_scc (cs_ci *A) ;
int cs_ci_scatter (const cs_ci *A, int j, double _Complex beta, int *w, 
    double _Complex *x, int mark,cs_ci *C, int nz) ;
int cs_ci_tdfs (int j, int k, int *head, const int *next, int *post,
    int *stack) ;

/* utilities */
cs_cid *cs_ci_dalloc (int m, int n) ;
cs_ci *cs_ci_done (cs_ci *C, void *w, void *x, int ok) ;
int *cs_ci_idone (int *p, cs_ci *C, void *w, int ok) ;
cs_cin *cs_ci_ndone (cs_cin *N, cs_ci *C, void *w, void *x, int ok) ;
cs_cid *cs_ci_ddone (cs_cid *D, cs_ci *C, void *w, int ok) ;

/* nearly private routines, only used in cs_ci_lu */
int cs_ci_reach (cs_ci *L, const cs_ci *B, int k, int *xi, const int *Pinv) ;
int cs_ci_splsolve (cs_ci *L, const cs_ci *B, int k, int *xi, 
    double _Complex *x, const int *Pinv) ;


/* -------------------------------------------------------------------------- */
/* complex/long version of CXSparse */
/* -------------------------------------------------------------------------- */

/* --- primary CSparse routines and data structures ------------------------- */

typedef struct cs_cl_sparse  /* matrix in compressed-column or triplet form */
{
    long nzmax ;    /* maximum number of entries */
    long m ;	    /* number of rows */
    long n ;	    /* number of columns */
    long *p ;	    /* column pointers (size n+1) or col indlces (size nzmax) */
    long *i ;	    /* row indices, size nzmax */
    double _Complex *x ;    /* numerical values, size nzmax */
    long nz ;	    /* # of entries in triplet matrix, -1 for compressed-col */
} cs_cl ;

cs_cl *cs_cl_add (const cs_cl *A, const cs_cl *B, double _Complex alpha,
    double _Complex beta) ;
long cs_cl_cholsol (const cs_cl *A, double _Complex *b, long order) ;
long cs_cl_dupl (cs_cl *A) ;
long cs_cl_entry (cs_cl *T, long i, long j, double _Complex x) ;
long cs_cl_lusol (const cs_cl *A, double _Complex *b, long order, double tol) ;
long cs_cl_gaxpy (const cs_cl *A, const double _Complex *x, double _Complex *y);
cs_cl *cs_cl_multiply (const cs_cl *A, const cs_cl *B) ;
long cs_cl_qrsol (const cs_cl *A, double _Complex *b, long order) ;
cs_cl *cs_cl_transpose (const cs_cl *A, long values) ;
cs_cl *cs_cl_triplet (const cs_cl *T) ;
double cs_cl_norm (const cs_cl *A) ;
long cs_cl_print (const cs_cl *A, long brief) ;
cs_cl *cs_cl_load (FILE *f) ;

/* utilities */
void *cs_cl_calloc (long n, size_t size) ;
void *cs_cl_free (void *p) ;
void *cs_cl_realloc (void *p, long n, size_t size, long *ok) ;
cs_cl *cs_cl_spalloc (long m, long n, long nzmax, long values, long t) ;
cs_cl *cs_cl_spfree (cs_cl *A) ;
long cs_cl_sprealloc (cs_cl *A, long nzmax) ;
void *cs_cl_malloc (long n, size_t size) ;

/* --- secondary CSparse routines and data structures ----------------------- */

typedef struct cs_cl_symbolic  /* symbolic Cholesky, LU, or QR analysis */
{
    long *Pinv ;    /* inverse row perm. for QR, fill red. perm for Chol */
    long *Q ;	    /* fill-reducing column permutation for LU and QR */
    long *parent ;  /* elimination tree for Cholesky and QR */
    long *cp ;	    /* column pointers for Cholesky, row counts for QR */
    long m2 ;	    /* # of rows for QR, after adding fictitious rows */
    long lnz ;	    /* # entries in L for LU or Cholesky; in V for QR */
    long unz ;	    /* # entries in U for LU; in R for QR */
} cs_cls ;

typedef struct cs_cl_numeric   /* numeric Cholesky, LU, or QR factorization */
{
    cs_cl *L ;	    /* L for LU and Cholesky, V for QR */
    cs_cl *U ;	    /* U for LU, R for QR, not used for Cholesky */
    long *Pinv ;    /* partial pivoting for LU */
    double _Complex *B ;	    /* beta [0..n-1] for QR */
} cs_cln ;

typedef struct cs_cl_dmperm_results    /* cs_cl_dmperm or cs_cl_scc output */
{
    long *P ;	    /* size m, row permutation */
    long *Q ;	    /* size n, column permutation */
    long *R ;	    /* size nb+1, block k is rows R[k] to R[k+1]-1 in A(P,Q) */
    long *S ;	    /* size nb+1, block k is cols S[k] to S[k+1]-1 in A(P,Q) */
    long nb ;	    /* # of blocks in fine dmperm decomposition */
    long rr [5] ;   /* coarse row decomposition */
    long cc [5] ;   /* coarse column decomposition */
} cs_cld ;

long *cs_cl_amd (const cs_cl *A, long order) ;
cs_cln *cs_cl_chol (const cs_cl *A, const cs_cls *S) ;
cs_cld *cs_cl_dmperm (const cs_cl *A) ;
long cs_cl_droptol (cs_cl *A, double tol) ;
long cs_cl_dropzeros (cs_cl *A) ;
long cs_cl_happly (const cs_cl *V, long i, double _Complex beta,
    double _Complex *x) ;
long cs_cl_ipvec (long n, const long *P, const double _Complex *b,
    double _Complex *x) ;
long cs_cl_lsolve (const cs_cl *L, double _Complex *x) ;
long cs_cl_ltsolve (const cs_cl *L, double _Complex *x) ;
cs_cln *cs_cl_lu (const cs_cl *A, const cs_cls *S, double tol) ;
cs_cl *cs_cl_permute (const cs_cl *A, const long *P, const long *Q,
    long values) ;
long *cs_cl_pinv (const long *P, long n) ;
long cs_cl_pvec (long n, const long *p, const double _Complex *b,
    double _Complex *x) ;
cs_cln *cs_cl_qr (const cs_cl *A, const cs_cls *S) ;
cs_cls *cs_cl_schol (const cs_cl *A, long order) ;
cs_cls *cs_cl_sqr (const cs_cl *A, long order, long qr) ;
cs_cl *cs_cl_symperm (const cs_cl *A, const long *Pinv, long values) ;
long cs_cl_usolve (const cs_cl *U, double _Complex *x) ;
long cs_cl_utsolve (const cs_cl *U, double _Complex *x) ;
long cs_cl_updown (cs_cl *L, long sigma, const cs_cl *C, const long *parent) ;

/* utilities */
cs_cls *cs_cl_sfree (cs_cls *S) ;
cs_cln *cs_cl_nfree (cs_cln *N) ;
cs_cld *cs_cl_dfree (cs_cld *D) ;

/* --- tertiary CSparse routines -------------------------------------------- */

long *cs_cl_counts (const cs_cl *A, const long *parent, const long *post,
    long ata) ;
long cs_cl_cumsum (long *p, long *c, long n) ;
long cs_cl_dfs (long j, cs_cl *L, long top, long *xi, long *pstack,
    const long *Pinv) ;
long *cs_cl_etree (const cs_cl *A, long ata) ;
long cs_cl_fkeep (cs_cl *A, long (*fkeep) (long, long, double _Complex, void *),
    void *other) ;
double _Complex cs_cl_house (double _Complex *x, double _Complex *beta, long n);
long *cs_cl_maxtrans (const cs_cl *A) ;
long *cs_cl_post (long n, const long *parent) ;
cs_cld *cs_cl_scc (cs_cl *A) ;
long cs_cl_scatter (const cs_cl *A, long j, double _Complex beta, long *w,
    double _Complex *x, long mark,cs_cl *C, long nz) ;
long cs_cl_tdfs (long j, long k, long *head, const long *next, long *post,
    long *stack) ;

/* utilities */
cs_cld *cs_cl_dalloc (long m, long n) ;
cs_cl *cs_cl_done (cs_cl *C, void *w, void *x, long ok) ;
long *cs_cl_idone (long *p, cs_cl *C, void *w, long ok) ;
cs_cln *cs_cl_ndone (cs_cln *N, cs_cl *C, void *w, void *x, long ok) ;
cs_cld *cs_cl_ddone (cs_cld *D, cs_cl *C, void *w, long ok) ;

/* nearly private routines, only used in cs_cl_lu */
long cs_cl_reach (cs_cl *L, const cs_cl *B, long k, long *xi,
    const long *Pinv) ;
long cs_cl_splsolve (cs_cl *L, const cs_cl *B, long k, long *xi, 
    double _Complex *x, const long *Pinv) ;


/* -------------------------------------------------------------------------- */
/* Macros for constructing each version of CSparse */
/* -------------------------------------------------------------------------- */

#ifdef CS_LONG
#define CS_INT long
#define CS_INT_MAX LONG_MAX
#define CS_ID "%ld"
#ifdef CS_COMPLEX
#define CS_ENTRY double _Complex
#define CS_NAME(nm) cs_cl ## nm
#else
#define CS_ENTRY double
#define CS_NAME(nm) cs_dl ## nm
#endif
#else
#define CS_INT int
#define CS_INT_MAX INT_MAX
#define CS_ID "%d"
#ifdef CS_COMPLEX
#define CS_ENTRY double _Complex
#define CS_NAME(nm) cs_ci ## nm
#else
#define CS_ENTRY double
#define CS_NAME(nm) cs_di ## nm
#endif
#endif

#ifdef CS_COMPLEX
#define CS_REAL(x) creal(x)
#define CS_IMAG(x) cimag(x)
#define CS_CONJ(x) conj(x)
#define CS_ABS(x) cabs(x)
#else
#define CS_REAL(x) (x)
#define CS_IMAG(x) (0.)
#define CS_CONJ(x) (x)
#define CS_ABS(x) fabs(x)
#endif

#define CS_MAX(a,b) (((a) > (b)) ? (a) : (b))
#define CS_MIN(a,b) (((a) < (b)) ? (a) : (b))
#define CS_FLIP(i) (-(i)-2)
#define CS_UNFLIP(i) (((i) < 0) ? CS_FLIP(i) : (i))
#define CS_MARKED(Ap,j) (Ap [j] < 0)
#define CS_MARK(Ap,j) { Ap [j] = CS_FLIP (Ap [j]) ; }

#define CS_OVERFLOW(n,size) (n > CS_INT_MAX / (CS_INT) size)

/* --- primary CSparse routines and data structures ------------------------- */

#define cs CS_NAME ()

#define cs_add CS_NAME (_add)
#define cs_cholsol CS_NAME (_cholsol)
#define cs_dupl CS_NAME (_dupl)
#define cs_entry CS_NAME (_entry)
#define cs_lusol CS_NAME (_lusol)
#define cs_gaxpy CS_NAME (_gaxpy)
#define cs_multiply CS_NAME (_multiply)
#define cs_qrsol CS_NAME (_qrsol)
#define cs_transpose CS_NAME (_transpose)
#define cs_triplet CS_NAME (_triplet)
#define cs_norm CS_NAME (_norm)
#define cs_print CS_NAME (_print)
#define cs_load CS_NAME (_load)

/* utilities */
#define cs_calloc CS_NAME (_calloc)
#define cs_free CS_NAME (_free)
#define cs_realloc CS_NAME (_realloc)
#define cs_spalloc CS_NAME (_spalloc)
#define cs_spfree CS_NAME (_spfree)
#define cs_sprealloc CS_NAME (_sprealloc)
#define cs_malloc CS_NAME (_malloc)

/* --- secondary CSparse routines and data structures ----------------------- */
#define css CS_NAME (s)
#define csn CS_NAME (n)
#define csd CS_NAME (d)

#define cs_amd CS_NAME (_amd)
#define cs_chol CS_NAME (_chol)
#define cs_dmperm CS_NAME (_dmperm)
#define cs_droptol CS_NAME (_droptol)
#define cs_dropzeros CS_NAME (_dropzeros)
#define cs_happly CS_NAME (_happly)
#define cs_ipvec CS_NAME (_ipvec)
#define cs_lsolve CS_NAME (_lsolve)
#define cs_ltsolve CS_NAME (_ltsolve)
#define cs_lu CS_NAME (_lu)
#define cs_permute CS_NAME (_permute)
#define cs_pinv CS_NAME (_pinv)
#define cs_pvec CS_NAME (_pvec)
#define cs_qr CS_NAME (_qr)
#define cs_schol CS_NAME (_schol)
#define cs_sqr CS_NAME (_sqr)
#define cs_symperm CS_NAME (_symperm)
#define cs_usolve CS_NAME (_usolve)
#define cs_utsolve CS_NAME (_utsolve)
#define cs_updown CS_NAME (_updown)

/* utilities */
#define cs_sfree CS_NAME (_sfree)
#define cs_nfree CS_NAME (_nfree)
#define cs_dfree CS_NAME (_dfree)

/* --- tertiary CSparse routines -------------------------------------------- */
#define cs_counts CS_NAME (_counts)
#define cs_cumsum CS_NAME (_cumsum)
#define cs_dfs CS_NAME (_dfs)
#define cs_etree CS_NAME (_etree)
#define cs_fkeep CS_NAME (_fkeep)
#define cs_house CS_NAME (_house)
#define cs_invmatch CS_NAME (_invmatch)
#define cs_maxtrans CS_NAME (_maxtrans)
#define cs_post CS_NAME (_post)
#define cs_scc CS_NAME (_scc)
#define cs_scatter CS_NAME (_scatter)
#define cs_tdfs CS_NAME (_tdfs)

/* utilities */
#define cs_dalloc CS_NAME (_dalloc)
#define cs_done CS_NAME (_done)
#define cs_idone CS_NAME (_idone)
#define cs_ndone CS_NAME (_ndone)
#define cs_ddone CS_NAME (_ddone)

/* nearly private routines */
#define cs_reach CS_NAME (_reach)
#define cs_splsolve CS_NAME (_splsolve)

/* -------------------------------------------------------------------------- */
/* Conversion routines */
/* -------------------------------------------------------------------------- */

cs_di *cs_i_real (cs_ci *A, int real) ;
cs_ci *cs_i_complex (cs_di *A, int real) ;
cs_dl *cs_l_real (cs_cl *A, long real) ;
cs_cl *cs_l_complex (cs_dl *A, long real) ;

#ifdef __cplusplus
}
#endif
#endif
